from contextlib import contextmanager

from .connection import db
from .models import context_auto_commit


@contextmanager
def transaction():
    token = context_auto_commit.set(False)
    yield
    context_auto_commit.reset(token)
    db.commit()



from pydantic import BaseSettings, validator
from typing import Any, Dict, List, Optional, Union

from bali.core import initialize


class Settings(BaseSettings):
    SERVER_NAME: str = 'bali-test'

    DATABASE_SERVER: str = '127.0.0.1'
    DATABASE_USER: str = 'root'
    DATABASE_PASSWORD: str = '123456'
    DATABASE_NAME: str = 'bali_test'
    SQLALCHEMY_DATABASE_URI: str = None

    @validator("SQLALCHEMY_DATABASE_URI", pre=True)
    def assemble_db_connection(cls, v: Optional[str], values: Dict[str, Any]) -> Any:
        if isinstance(v, str):
            return v
        return 'mysql+pymysql://{}:{}@{}/{}'.format(
            values.get("DATABASE_USER"), values.get("DATABASE_PASSWORD"),
            values.get("DATABASE_SERVER"), values.get("DATABASE_NAME")
        )

    CACHE_ADDRESS: str = '127.0.0.1'
    CACHE_PASSWORD: str = '123456'


settings = Settings()

initialize(settings)



"""
gRPC interceptors
"""
import grpc
import logging

from google.protobuf import json_format

from bali.core import db

from typing import Callable, Any
from grpc_interceptor import ServerInterceptor
from .core import _settings

logger = logging.getLogger('bali')


class ProcessInterceptor(ServerInterceptor):
    def setup(self):
        pass

    def teardown(self):
        try:
            db.remove()
        except Exception:
            pass

    @staticmethod
    def log(request, method_name, log_type):
        logger.info(
            '%s %s: %s',
            method_name,
            json_format.MessageToDict(
                request, including_default_value_fields=True, preserving_proto_field_name=True
            ),
            log_type,
        )

    def intercept(
        self,
        method: Callable,
        request: Any,
        context: grpc.ServicerContext,
        method_name: str,
    ) -> Any:
        """
        
        :param method:
        :param request:
        :param context:
        :param method_name:
        :return:
        """
        self.setup()
        try:
            _settings.ENABLED_RPC_LOGGING and self.log(request, method_name, 'Request')
            result = method(request, context)
            _settings.ENABLED_RPC_LOGGING and self.log(request, method_name, 'Response')
        finally:
            self.teardown()

        return result



from typing import Tuple

from sqlalchemy import desc, asc
from sqlalchemy.sql import operators
from sqlalchemy.sql.expression import extract

from ..exceptions import OperatorModelError

OPERATOR_SPLITTER = '__'

OPERATORS = {
    'isnull': lambda c, v: (c == None) if v else (c != None),  # noqa
    'exact': operators.eq,
    'ne': operators.ne,  # not equal or is not (for None)
    'gt': operators.gt,
    'gte': operators.ge,
    'lt': operators.lt,
    'lte': operators.le,
    'in': operators.in_op,
    'notin': operators.notin_op,
    'between': lambda c, v: c.between(v[0], v[1]),
    'like': operators.like_op,
    'ilike': operators.ilike_op,
    'startswith': operators.startswith_op,
    'istartswith': lambda c, v: c.ilike(v + '%'),
    'endswith': operators.endswith_op,
    'iendswith': lambda c, v: c.ilike('%' + v),
    'contains': lambda c, v: c.ilike('%{v}%'.format(v=v)),
    'year': lambda c, v: extract('year', c) == v,
    'year_ne': lambda c, v: extract('year', c) != v,
    'year_gt': lambda c, v: extract('year', c) > v,
    'year_ge': lambda c, v: extract('year', c) >= v,
    'year_lt': lambda c, v: extract('year', c) < v,
    'year_le': lambda c, v: extract('year', c) <= v,
    'month': lambda c, v: extract('month', c) == v,
    'month_ne': lambda c, v: extract('month', c) != v,
    'month_gt': lambda c, v: extract('month', c) > v,
    'month_ge': lambda c, v: extract('month', c) >= v,
    'month_lt': lambda c, v: extract('month', c) < v,
    'month_le': lambda c, v: extract('month', c) <= v,
    'day': lambda c, v: extract('day', c) == v,
    'day_ne': lambda c, v: extract('day', c) != v,
    'day_gt': lambda c, v: extract('day', c) > v,
    'day_ge': lambda c, v: extract('day', c) >= v,
    'day_lt': lambda c, v: extract('day', c) < v,
    'day_le': lambda c, v: extract('day', c) <= v,
}


def get_filters_expr(cls, **filters):
    if not callable(cls):
        raise OperatorModelError('Operator model must provide')

    expressions = []
    for attr, value in filters.items():
        if OPERATOR_SPLITTER in attr:
            attr_name, op_name = attr.rsplit(OPERATOR_SPLITTER, 1)
            if op_name not in OPERATORS:
                raise KeyError(
                    'Expression `{}` has incorrect '
                    'operator `{}`'.format(attr, op_name)
                )
            op = OPERATORS[op_name]
        else:
            attr_name, op = attr, operators.eq

        column = getattr(cls, attr_name)
        expressions.append(op(column, value))

    return expressions



from datetime import datetime

from bali.utils import timezone, dateparser, get_beginning_datetime


def test_parse_dmy_date():
    s = '04/03/2014 22:07:40'
    expect = datetime(2014, 3, 4, 22, 7, 40)
    assert dateparser.parse(s, settings={'DATE_ORDER': 'DMY'}) == expect


def test_get_year_beginning_datetime():
    d = get_beginning_datetime(year=2000)
    assert timezone.make_naive(d) == datetime(2000, 1, 1)


def test_get_month_beginning_datetime():
    d = get_beginning_datetime(year=2000, month=2)
    assert timezone.make_naive(d) == datetime(2000, 2, 1)


def test_get_day_beginning_datetime():
    d = get_beginning_datetime(year=2000, month=2, day=3)
    assert timezone.make_naive(d) == datetime(2000, 2, 3)



import hashlib
import os
import pickle
from urllib.parse import quote

from redis import StrictRedis

from bali.core import cache, cache_memoize

HOST = '127.0.0.1'
PREFIX = 'test_service'


def noop(*args):
    return args


args_rewrite = noop


def _default_make_cache_key(*args, prefix_=None, **kwargs):
    cache_key = "cache_memoize:" + ":" + prefix_ + ":".join(
        [quote(str(x)) for x in args_rewrite(args)] +
        [quote("{}={}".format(k, v)) for k, v in kwargs.items()]
    )
    return hashlib.md5(cache_key.encode('utf-8')).hexdigest()


@cache_memoize(30)
def foo():
    return '0'


@cache_memoize(30)
def foo_with_args(a):
    return f'0{a}'


class A:
    @cache_memoize(10)
    def foo(self):
        return 'A'

    @cache_memoize(10)
    def foo_with_args(self, a):
        return f'A{a}'

    @staticmethod
    @cache_memoize(10)
    def foo_with_args_s(a):
        return f'A{a}s'


class B:
    @cache_memoize(10)
    def foo(self):
        return 'B'

    @cache_memoize(10)
    def foo_with_args(self, a):
        return f'B{a}'

    @staticmethod
    @cache_memoize(10)
    def foo_with_args_s(a):
        return f'B{a}s'


class TestCacheMemoize:
    # noinspection PyMethodMayBeStatic
    def setup_class(self):
        self.redis = StrictRedis(HOST, password=os.getenv('CACHE_PASSWORD'))
        cache.connect(HOST, password=os.getenv('CACHE_PASSWORD'), prefix=PREFIX)

    def teardown_class(self):
        pass

    def test_cache_into_redis(self):
        expect_value = foo()
        prefix_ = ".".join((foo.__module__ or "", foo.__qualname__))
        key = _default_make_cache_key(prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result = pickle.loads(self.redis.get(redis_key))
        assert result == expect_value, 'Cache into redis verify failed'

    def test_cache_same_function_name_not_conflict(self):
        """Same function name in different module should not conflict"""
        a = A()
        b = B()
        value_a = a.foo()
        value_b = b.foo()
        assert value_a == 'A'
        assert value_b == 'B'

        # value a in redis
        prefix_ = ".".join((a.foo.__module__ or "", a.foo.__qualname__))
        key = _default_make_cache_key(a, prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result_a = pickle.loads(self.redis.get(redis_key))

        # value b in redis
        prefix_ = ".".join((b.foo.__module__ or "", b.foo.__qualname__))
        key = _default_make_cache_key(b, prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result_b = pickle.loads(self.redis.get(redis_key))

        assert result_a == value_a
        assert result_b == value_b

    def test_cache_same_function_args_not_conflict(self):
        """Same function name with args in different module should not conflict"""
        a = A()
        b = B()
        value_a = a.foo_with_args('9')
        value_b = b.foo_with_args('9')
        assert value_a == 'A9'
        assert value_b == 'B9'

        # value a in redis
        prefix_ = ".".join((a.foo_with_args.__module__ or "", a.foo_with_args.__qualname__))
        key = _default_make_cache_key(a, '9', prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result_a = pickle.loads(self.redis.get(redis_key))

        # value b in redis
        prefix_ = ".".join((b.foo_with_args.__module__ or "", b.foo_with_args.__qualname__))
        key = _default_make_cache_key(b, '9', prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result_b = pickle.loads(self.redis.get(redis_key))

        assert result_a == value_a
        assert result_b == value_b

    def test_cache_same_static_function_name_not_conflict(self):
        """Same static function name in different module should not conflict"""
        a = A()
        b = B()
        value_a = a.foo_with_args_s('1')
        value_b = b.foo_with_args_s('1')
        assert value_a == 'A1s'
        assert value_b == 'B1s'

        # value a in redis
        prefix_ = ".".join((a.foo_with_args_s.__module__ or "", a.foo_with_args_s.__qualname__))
        key = _default_make_cache_key('1', prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result_a = pickle.loads(self.redis.get(redis_key))

        # value b in redis
        prefix_ = ".".join((b.foo_with_args_s.__module__ or "", b.foo_with_args_s.__qualname__))
        key = _default_make_cache_key('1', prefix_=prefix_)
        redis_key = f'{PREFIX}:cache:{key}'
        result_b = pickle.loads(self.redis.get(redis_key))

        assert result_a == value_a
        assert result_b == value_b



import os
import pickle

import pytest

from bali.core import cache
from redis import StrictRedis

HOST = '127.0.0.1'
PREFIX = 'test_service'


class TestCache:
    # noinspection PyMethodMayBeStatic
    def setup_class(self):
        cache.connect(HOST, password=os.getenv('CACHE_PASSWORD'), prefix=PREFIX)

    def teardown_class(self):
        pass

    def test_cache_int(self):
        key = 'test_int'
        value = 1
        result = cache.set('test_int', value)
        assert result is True, 'Cache int value failed'

        redis = StrictRedis(HOST, password=os.getenv('CACHE_PASSWORD'))
        redis_key = f'{PREFIX}:cache:{key}'
        assert int(redis.get(redis_key).decode('utf-8')) == value, 'Cache int value verify failed'

        result = cache.get(key)
        assert result is value, 'Cache int value get failed'

    def test_cache_str(self):
        key = 'test_str'
        value = 'hello, bali'
        result = cache.set(key, value)
        assert result is True, 'Cache str value failed'

        redis = StrictRedis(HOST, password=os.getenv('CACHE_PASSWORD'))
        redis_key = f'{PREFIX}:cache:{key}'
        redis_result = pickle.loads(redis.get(redis_key))
        assert redis_result == value, 'Cache str value verify failed'

        result = cache.get(key)
        assert result == value, 'Cache str value get failed'

    def test_cache_list(self):
        key = 'test_list'
        value = [1, 2, 3, 'a', (1, 'zzz')]
        result = cache.set(key, value)
        assert result is True, 'Cache list value failed'

        redis = StrictRedis(HOST, password=os.getenv('CACHE_PASSWORD'))
        redis_key = f'{PREFIX}:cache:{key}'
        redis_result = pickle.loads(redis.get(redis_key))
        assert redis_result == value, 'Cache list value verify failed'

        result = cache.get(key)
        assert result == value, 'Cache list value get failed'

    def test_cache_dict(self):
        key = 'test_dict'
        value = {'test': 1, 'test2': 'test', 'test3': {'aa': 1}}
        result = cache.set(key, value)
        assert result is True, 'Cache dict value failed'

        redis = StrictRedis(HOST, password=os.getenv('CACHE_PASSWORD'))
        redis_key = f'{PREFIX}:cache:{key}'
        redis_result = pickle.loads(redis.get(redis_key))
        assert redis_result == value, 'Cache dict value verify failed'

        result = cache.get(key)
        assert result == value, 'Cache dict value get failed'

    def test_not_exists(self):
        result = cache.get('no-exist-key')
        assert result is None, 'Should return none when cache not exists'

    def test_incr_raise_ex(self):
        """
        incr a non exist key will raise ValueError exception
        """
        with pytest.raises(ValueError):
            cache.incr('no-exist-key')

    def test_incr(self):
        key = 'test_incr'
        cache.set(key, 0)
        next_value = cache.incr(key)
        assert next_value == 1
        next_value = cache.incr(key)
        assert next_value == 2
        next_value = cache.incr(key, 2)
        assert next_value == 4

    def test_decr(self):
        key = 'test_decr'
        cache.set(key, 0)
        next_value = cache.decr(key)
        assert next_value == -1
        next_value = cache.decr(key)
        assert next_value == -2
        next_value = cache.decr(key, 2)
        assert next_value == -4



import functools

from fastapi.dependencies.utils import get_typed_signature
from fastapi_pagination import LimitOffsetParams, set_page
from fastapi_pagination.limit_offset import Page
from pydantic import BaseModel

from .exceptions import ReturnTypeError
from .paginate import paginate
from .utils import MessageToDict, ParseDict


def compatible_method(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):

        # Put args to inner function from request object
        if self._is_rpc:
            request_data = MessageToDict(
                self._request,
                including_default_value_fields=True,
                preserving_proto_field_name=True,
            )

            if func.__name__ == 'get':
                pk = self._request.id
                result = func(self, pk)
                if not isinstance(result, dict):
                    result = result.dict()
                response_data = {'data': result}

            elif func.__name__ == 'list':
                schema_in = get_schema_in(func)
                result = func(self, schema_in(**request_data))
                # Paginated the result queryset or iterable object
                if isinstance(result, BaseModel):
                    raise ReturnTypeError('Generic actions `list` should return a sequence')
                else:
                    set_page(Page)
                    params = LimitOffsetParams(
                        limit=request_data.get('limit') or 10,
                        offset=request_data.get('offset'),
                    )
                    response_data = paginate(result, params=params, is_rpc=True)

            elif func.__name__ in ['create', 'update']:
                schema_in = get_schema_in(func)
                data = request_data.get('data')
                result = func(self, schema_in(**data))
                if not isinstance(result, dict):
                    result = result.dict()
                response_data = {'data': result}

            elif func.__name__ == 'delete':
                pk = self._request.id
                result = func(self, pk)
                response_data = {'result': bool(result)}

            else:
                # custom action
                schema_in = get_schema_in(func)
                result = func(self, schema_in(**request_data))
                if not isinstance(result, dict):
                    result = result.dict()
                response_data = result

            # Convert response data to gRPC response
            return ParseDict(response_data, self._response_message(), ignore_unknown_fields=True)

        return func(self, *args, **kwargs)

    return wrapper


def action(methods=None, detail=None, **kwargs):
    """
    Mark a Resource method as a routable action.

    Set the `detail` boolean to determine if this action should apply to
    instance/detail requests or collection/list requests.

    :param methods:
    :param detail:
    :param kwargs:
        http_only: the action only support http
        rpc_only: the action only support rpc
    """
    methods = ['get'] if (methods is None) else methods
    methods = [method.lower() for method in methods]

    http_only = kwargs.get('http_only', False)
    rpc_only = kwargs.get('rpc_only', False)

    class Action:
        def __init__(self, func):
            self.func = func

        def __set_name__(self, owner, name):
            # replace ourself with the original method
            setattr(owner, name, compatible_method(self.func))

            # Append actions to Resource._actions
            # Only for http actions
            if rpc_only:
                return

            try:
                schema_in_annotation = get_schema_in(self.func)
            except ValueError:
                schema_in_annotation = None

            _actions = getattr(owner, '_actions')
            _actions[self.func.__name__] = {
                'detail': detail,
                'methods': methods,
                'schema_in_annotation': schema_in_annotation,
            }
            setattr(owner, '_actions', _actions)

    return Action


def get_schema_in(func):
    typed_signature = get_typed_signature(func)
    signature_params = typed_signature.parameters

    # 1st argument is self
    # 2st argument is schema_in
    index = 0
    for param_name, param in signature_params.items():
        index += 1
        if index == 2 or param_name == 'schema_in':
            schema_in = param.annotation
            if not schema_in:
                raise ValueError('Custom actions must provide `schema_in` argument with annotation')

            return schema_in
    else:
        raise ValueError('Custom actions arguments error')



from concurrent import futures

import grpc

from ._utils import get_service_adder
from .interceptors import ProcessInterceptor


class GRPCTestBase:
    server = None
    server_class = None  # must provided in inherit class
    port = 50001

    service_target = 'localhost'  # grpc channel target address
    pb2 = None
    pb2_grpc = None

    def setup_class(self):
        self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=10),
            interceptors=[ProcessInterceptor()],
        )

        add_service_to_server = get_service_adder(self.pb2_grpc)
        add_service_to_server(self.server_class(), self.server)
        self.server.add_insecure_port(f'[::]:{self.port}')
        self.server.start()

        self.service_target = f'localhost:{self.port}'

    def teardown_class(self):
        self.server.stop(None)



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Python implementation of the GRPC helloworld.Greeter client."""

from __future__ import print_function

import logging

import grpc

import helloworld_pb2
import helloworld_pb2_grpc
from bali.utils import MessageToDict, ParseDict


def run():
    # NOTE(gRPC Python Team): .close() is possible on a channel and should be
    # used in circumstances in which the with statement does not fit the needs
    # of the code.
    with grpc.insecure_channel('localhost:50051') as channel:
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        response = stub.SayHello(helloworld_pb2.HelloRequest(name='you'))
        print("Greeter client received <SayHello>: %s" % response.message)

    with grpc.insecure_channel('localhost:50051') as channel:
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        response = stub.GetGreeter(helloworld_pb2.GetRequest(id=3))
        print("Greeter client received <GetGreeter>: %s" % MessageToDict(response))

    with grpc.insecure_channel('localhost:50051') as channel:
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        data = {
            'id': 1,
            'content': 'Greeter',
        }
        request_pb = ParseDict(
            {'data': data},
            helloworld_pb2.CreateRequest(),
            ignore_unknown_fields=True,
        )
        response = stub.CreateGreeter(request_pb)
        print("Greeter client received <CreateGreeter>: %s" % MessageToDict(response))

    with grpc.insecure_channel('localhost:50051') as channel:
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        response = stub.ListGreeter(helloworld_pb2.ListRequest(limit=2, offset=3))
        print("Greeter client received <ListGreeter>: %s" % MessageToDict(response))

    with grpc.insecure_channel('localhost:50051') as channel:
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        response = stub.GetItem(helloworld_pb2.GetRequest(id=1))
        print("Greeter client received <GetItem>: %s" % MessageToDict(response))

    with grpc.insecure_channel('localhost:50051') as channel:
        stub = helloworld_pb2_grpc.GreeterStub(channel)
        response = stub.ListItems(helloworld_pb2.ListRequest(limit=2, offset=3))
        print("Greeter client received <ListItems>: %s" % MessageToDict(response))


if __name__ == '__main__':
    logging.basicConfig()
    run()



"""
FastAPI middleware
"""
from fastapi import Request


async def process_middleware(request: Request, call_next):
    from bali.core import db

    try:
        response = await call_next(request)
    except Exception:
        raise
    else:
        return response
    finally:

        # remove db session when FastAPI request ended
        try:
            db.remove()
        except Exception:
            pass



from typing import Container, Optional, Type

from pydantic import BaseConfig, BaseModel, create_model
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.properties import ColumnProperty

from .generic import *


class OrmConfig(BaseConfig):
    orm_mode = True


def model_to_schema(
    db_model: Type,
    *,
    config: Type = OrmConfig,
    exclude: Container[str] = [],
    partial=False  # noqa
) -> Type[BaseModel]:
    mapper = inspect(db_model)
    fields = {}
    for attr in mapper.attrs:
        if isinstance(attr, ColumnProperty):
            if attr.columns:
                name = attr.key
                if name in exclude:
                    continue
                column = attr.columns[0]
                python_type: Optional[type] = None
                if hasattr(column.type, "impl"):
                    if hasattr(column.type.impl, "python_type"):
                        python_type = column.type.impl.python_type
                elif hasattr(column.type, "python_type"):
                    python_type = column.type.python_type
                assert python_type, f"Could not infer python_type for {column}"

                default = None
                if column.default is None and not column.nullable:
                    default = ...

                if partial:
                    fields[name] = (Optional[python_type], None)
                else:
                    fields[name] = (python_type, default)
    pydantic_model = create_model(
        db_model.__name__,
        __config__=config,
        **fields  # noqa
    )
    return pydantic_model



"""Bali Resource

A Resource layer base class to handle FastAPI and gRPC requests and responses

Resource's method input is Pydantic schema

"""
import inspect
import logging
import typing
from collections import OrderedDict
from typing import Optional, Callable

from fastapi import Request, HTTPException, status
from fastapi_pagination import LimitOffsetPage
from google.protobuf import message
from pydantic import BaseModel

from .paginate import paginate
from .routing import APIRouter
from .schemas import ListRequest, ResultResponse

__all__ = ['Resource']


class Resource:
    """Base Resource

    Generic Actions is get, list, create, update, delete
    """

    # Resource's name, should automatic recognition is not provided
    name = None

    schema = None
    filters = []
    permission_classes = []

    _actions = OrderedDict()

    def __init__(self, request=None, context=None, response_message=None):
        self._request = request
        self._context = context
        self._response_message = response_message

        self._is_rpc = isinstance(self._request, message.Message)
        self._is_http = not self._is_rpc

        self.auth = BaseModel()

    @classmethod
    def as_router(cls):
        return RouterGenerator(cls)()


# noinspection PyShadowingBuiltins,PyUnresolvedReferences,PyProtectedMember,PyShadowingNames,DuplicatedCode
class RouterGenerator:
    """
    Generator router from Resource class
    """
    def __init__(self, cls):
        self.cls = cls
        self.router = APIRouter()
        self._ordered_filters = self._get_ordered_filters()

    def __call__(self):

        generic_actions = ['get', 'list', 'create', 'update', 'delete']
        # To fixed generic get action `/item/{id}` conflict with custom action `/item/hello`,
        # must make sure get action `/item/{id}` is below custom action
        actions = sorted(self.cls._actions.keys(), key=lambda x: x in generic_actions)

        # noinspection PyProtectedMember
        for action in actions:
            extra = self.cls._actions.get(action)
            if not hasattr(self.cls, action):
                continue
            self.add_route(action, extra)
        return self.router

    @property
    def resource_name(self):
        return self.cls.__name__.replace('Resource', '')

    @property
    def primary_key(self):
        return 'id'

    def check_permissions(self, resource):
        for permission_class in self.cls.permission_classes:
            permission = permission_class(resource)
            if not permission.check():
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail='Permission Denied',
                )

    def _get_ordered_filters(self):
        filters = OrderedDict()
        for item in self.cls.filters:
            k = item.keys().__iter__().__next__()
            v = item.values().__iter__().__next__()
            filters[k] = v
        return filters

    def _list(self) -> Callable:
        """
        list action default using fastapi-pagination to process paginate
        """
        resource = self.cls()

        def route(request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            params = request.query_params._dict

            # parse filters
            filters = {}
            for k, v in params.items():
                if k not in self._ordered_filters:
                    continue
                # noinspection PyBroadException
                try:
                    # Convert param and retrieve param
                    params_converter = self._ordered_filters.get(k)
                    if isinstance(params_converter, typing._GenericAlias):
                        params_converter = params_converter.__args__[0]
                    filters[k] = params_converter(v)
                except Exception as ex:
                    logging.warning(
                        'Query params `%s`(value: %s) type convert failed, exception: %s', k, v, ex
                    )
                    continue

            schema_in = ListRequest(**params, filters=filters)

            result = getattr(resource, 'list')(schema_in)
            return paginate(result)

        # update route's signatures
        parameters = []
        for k, v in self._ordered_filters.items():
            default = inspect.Parameter.empty
            if isinstance(v, typing._GenericAlias):
                default = None if type(None) in v.__args__ else default
            parameters.append(
                inspect.Parameter(
                    name=k,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=default,
                    annotation=v,
                )
            )
        sig = inspect.signature(route)
        parameters.extend(sig.parameters.values())
        route.__signature__ = sig.replace(parameters=parameters)

        return route

    def _get(self) -> Callable:
        resource = self.cls()

        def route(id: int, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return getattr(resource, 'get')(pk=id)

        return route

    def _create(self) -> Callable:
        resource = self.cls()

        def route(schema_in: resource.schema, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return getattr(resource, 'create')(schema_in)

        return route

    def _update(self) -> Callable:
        resource = self.cls()

        def route(schema_in: resource.schema, id: int, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return getattr(resource, 'update')(schema_in, pk=id)

        return route

    def _delete(self) -> Callable:
        resource = self.cls()

        def route(id: int, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return getattr(resource, 'delete')(pk=id)

        return route

    def get_endpoint(self, action, detail=False, methods=list, schema_in_annotation=None):
        """Convert Resource instance method to FastAPI endpoint"""
        resource = self.cls()

        def endpoint(request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return getattr(resource, action)()

        def endpoint_detail(pk: int, request: Request = None):
            resource._request = request
            self.check_permissions(resource)
            return getattr(resource, action)(pk)

        def endpoint_schema(schema_in: BaseModel = None, request: Request = None, **kwargs):
            resource._request = request
            self.check_permissions(resource)
            if 'get' in methods and schema_in_annotation:
                schema_in = schema_in_annotation(**kwargs)
            return getattr(resource, action)(schema_in)

        sig = inspect.signature(getattr(self.cls, action))

        route = endpoint
        if detail:
            route = endpoint_detail
        elif 'schema_in' in sig.parameters:
            route = endpoint_schema
            params = list(sig.parameters.values())[1:]
            if 'get' in methods and schema_in_annotation:
                # Destructor the `schema_in` to Query
                for field, annotation in schema_in_annotation.__fields__.items():
                    params = params[1:]
                    params.append(
                        inspect.Parameter(
                            name=field,
                            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                            default=annotation.default,
                            annotation=annotation.type_,
                        )
                    )
                route.__signature__ = sig.replace(parameters=params)
            else:
                params = list(sig.parameters.values())[1:]

            params.append(
                inspect.Parameter(
                    name='request',
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=None,
                    annotation=Request,
                )
            )
            route.__signature__ = sig.replace(parameters=params)

        return route

    def add_route(self, action, extra):
        if action == 'list':
            self.router.add_api_route(
                '',
                self._list(),
                methods=['GET'],
                response_model=LimitOffsetPage[self.cls.schema],
                summary=f'List {self.resource_name}'
            )
        elif action == 'create':
            self.router.add_api_route(
                '',
                self._create(),
                methods=['POST'],
                response_model=self.cls.schema and Optional[self.cls.schema],
                summary=f'Create {self.resource_name}'
            )
        elif action == 'get':
            self.router.add_api_route(
                '/{%s}' % self.primary_key,
                self._get(),
                methods=['GET'],
                response_model=self.cls.schema,
                summary=f'Get {self.resource_name}'
            )
        elif action == 'update':
            self.router.add_api_route(
                '/{%s}' % self.primary_key,
                self._update(),
                methods=['PATCH'],
                response_model=self.cls.schema,
                summary=f'Update {self.resource_name}'
            )
        elif action == 'delete':
            self.router.add_api_route(
                '/{%s}' % self.primary_key,
                self._delete(),
                methods=['DELETE'],
                response_model=ResultResponse,
                summary=f'Delete {self.resource_name}'
            )
        else:
            detail = extra.get('detail')
            methods = extra.get('methods')
            schema_in_annotation = extra.get('schema_in_annotation')
            path = '/{%s}' % self.primary_key if detail else ''
            self.router.add_api_route(
                f"{path}/{action.replace('_', '-')}",
                self.get_endpoint(
                    action,
                    detail,
                    methods=methods,
                    schema_in_annotation=schema_in_annotation,
                ),
                methods=methods,
                summary=f"{action.replace('_', ' ')}"
            )



from sqlalchemy import Column, Integer, String

from bali.db import db
from bali.db.operators import get_filters_expr

DB_URI = 'sqlite:///:memory:'


def test_db_operators_fetch_rows():
    db.connect(DB_URI)

    class User(db.BaseModel):
        __tablename__ = "users"
        id = Column(Integer, primary_key=True)
        username = Column(String(50), default='')
        age = Column(Integer)

    db.create_all()
    lucy = User.create(**{
        'username': 'Lucy',
        'age': 13,
    })
    jack = User.create(**{
        'username': 'Jack',
        'age': 20,
    })
    groot = User.create(**{
        'username': 'Groot',
        'age': 32,
    })

    assert lucy.id > 0

    users = User.query().filter(User.username.like('%c%'), User.age > 0).all()
    assert len(users) == 2, 'Fetch users count in common way should work properly'

    filters = {
        'username__like': '%c%',
        'age__gt': 0,
    }
    users = User.query().filter(*get_filters_expr(User, **filters)).all()
    assert len(users) == 2



"""Bali Cache

Support redis backend 
"""
import pickle
from redis import StrictRedis

DEFAULT_TIMEOUT = 300


class Cache:
    def __init__(self):
        self.configured = False
        self.host = None
        self.port = None
        self.password = None
        self.prefix = None

        self._client = None

    def _connect(self):
        self._client = StrictRedis(self.host, port=self.port, password=self.password)

    def connect(self, host, port=6379, password=None, prefix=None):
        """
        Only support db 0
        """
        self.configured = True
        self.host = host
        self.port = port
        self.password = password
        self.prefix = prefix

        if self.prefix is None:
            raise Exception('Prefix should provided')

        self._connect()

    def __getattribute__(self, attr, *args, **kwargs):
        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if not self._client:
                raise Exception('Cache client not initialized')

            return getattr(self._client, attr)

    def make_key(self, key):
        return f'{self.prefix}:cache:{key}'

    @staticmethod
    def encode(value):
        if isinstance(value, bool) or not isinstance(value, int):
            return pickle.dumps(value)
        return value

    @staticmethod
    def decode(value):
        try:
            value = int(value)
        except (ValueError, TypeError):
            value = pickle.loads(value)
        return value

    def set(self, key, value, timeout=DEFAULT_TIMEOUT):
        key = self.make_key(key)
        value = self.encode(value)
        return bool(self._client.set(key, value, ex=timeout))

    def get(self, key):
        key = self.make_key(key)
        value = self._client.get(key)
        if value is None:
            return value
        return self.decode(value)

    def _incr(self, key, delta=1):
        """Always check key exists
        """
        # if key expired after exists check, then we get
        # key with wrong value and ttl -1.
        # use lua script for atomicity
        lua = """
        local exists = redis.call('EXISTS', KEYS[1])
        if (exists == 1) then
            return redis.call('INCRBY', KEYS[1], ARGV[1])
        else return false end
        """
        value = self._client.eval(lua, 1, key, delta)
        if value is None:
            raise ValueError("Key '%s' not found" % key)

        return value

    def incr(self, key, delta=1):
        key = self.make_key(key)
        return self._incr(key, delta=delta)

    def decr(self, key, delta=1):
        key = self.make_key(key)
        return self._incr(key, delta=-delta)


cache = Cache()



from pydantic import BaseModel
from typing import List, Optional

from bali.decorators import action
from bali.resource import Resource
from bali.schemas import GetRequest, ListRequest
from permissions import IsAuthenticated

GREETERS = [{'id': i, 'content': 'Hi, number %s' % i} for i in range(10)]


class Greeter(BaseModel):
    id: int
    content: str


class GreeterFilter(BaseModel):
    ids: List[int]


class GreeterResource(Resource):

    schema = Greeter
    filters = [
        {'name': str},
        {'title': Optional[str]},
    ]
    permission_classes = [IsAuthenticated]

    @action()
    def get(self, pk=None):
        return [g for g in GREETERS if g.get('id') == pk][0]

    @action()
    def list(self, schema_in: ListRequest = None):
        # `list` NOT FULL SUPPORT HTTP REQUEST
        # return GREETERS[:schema_in.limit]
        print(schema_in.filters.get('name'))
        return GREETERS

    @action()
    def create(self, schema_in: schema = None):
        return {'id': schema_in.id, 'content': schema_in.content}

    @action()
    def update(self, schema_in: schema = None, pk=None):
        return {'id': pk, 'content': schema_in.content}

    @action()
    def delete(self, pk=None):
        return {'result': True}

    @action(detail=False)
    def recents(self):
        return GREETERS[:2]

    @action(detail=True)
    def items_recents(self, pk=None):
        return [g for g in GREETERS if g.get('id') == pk]

    @action(detail=False, methods=['post'])
    def custom_create(self, schema_in: GreeterFilter):
        print('schema_in', schema_in)
        return GREETERS[0]



import bali
from setuptools import find_packages, setup


def read(f):
    return open(f, 'r', encoding='utf-8').read()


INSTALL_REQUIREMENTS = [
    'dateparser==1.0.0',
    'decamelize==0.1.2',
    'fastapi[all]==0.63.0',
    'fastapi-pagination==0.7.0',
    'grpcio==1.32.0',
    'grpcio-tools==1.32.0',
    'grpc-interceptor==0.13.0',
    'PyMySQL==0.10.1',
    'passlib[bcrypt]==1.7.2',
    'pillow==7.2.0',
    'protobuf==3.13.0',
    'python-jose[cryptography]==3.2.0',
    'pydantic-sqlalchemy==0.0.7',
    'redis==3.5.3',
    'SQLAlchemy==1.3.19',
    'sqla-wrapper==4.200628',
    'typer==0.3.2',
    'uvicorn==0.12.3',
]

setup(
    name='bali-core',
    version=bali.__version__,
    description='Simplify FastAPI integrate gRPC services development',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/JoshYuJump/bali',
    author='Josh.Yu',
    author_email='josh.yu_8@live.com',
    license='MIT',
    install_requires=INSTALL_REQUIREMENTS,
    packages=find_packages(),
    package_data={'bali': ['db/*.pyi']},
    include_package_data=True,
    zip_safe=False,
)



from sqlalchemy import Column, Integer

from bali.db import db
from bali.schema import model_to_schema

DB_URI = 'sqlite:///:memory:'


def test_model_to_schema():
    db.connect(DB_URI)

    class User(db.BaseModel):
        __tablename__ = "users"
        id = Column(Integer, primary_key=True)

    db.create_all()

    # noinspection PyPep8Naming
    UserSchema = model_to_schema(User)

    user_schema = UserSchema(id=1)
    assert user_schema.id == 1


# noinspection PyUnresolvedReferences
def test_model_to_schema_from_orm():
    db.connect(DB_URI)

    class User(db.BaseModel):
        __tablename__ = "users"
        id = Column(Integer, primary_key=True)

    db.create_all()
    user = User.create()
    # noinspection PyPep8Naming
    UserSchema = model_to_schema(User)

    user_schema = UserSchema.from_orm(user)
    assert user_schema.id == 1
    assert user_schema.created_time <= user_schema.updated_time
    assert user_schema.is_active

    expected = {
        'id': user.id,
        'created_time': user.created_time,
        'updated_time': user.updated_time,
        'is_active': user.is_active,
    }

    assert user_schema.dict() == expected, 'Schema to dict value should equal origin model'





from fastapi_pagination import LimitOffsetPage, LimitOffsetParams, paginate
from fastapi import Query

from bali.core import APIRouter, cache_memoize
from models import Item
from resources import GreeterResource, ItemResource
from schemas import ItemModel

router = APIRouter()


@cache_memoize(10)
def get_cacheable_items():
    items = list(Item.query().all())
    print('load from db')
    return items


@router.get("/")
async def root():
    return {"message": "Hello World"}


@router.get("/items", response_model=LimitOffsetPage[ItemModel])
def list_items():
    return paginate(get_cacheable_items())


hello_greeter_router = GreeterResource.as_router()
item_router = ItemResource.as_router()
router.include_router(hello_greeter_router, prefix='/hello-greeters')
router.include_router(item_router, prefix='/hello-items')



from sqlalchemy import Column, Integer

from bali.db import db

DB_URI = 'sqlite:///:memory:'


# noinspection PyProtectedMember
def test_db_connect():
    db.connect(DB_URI)
    assert db._session is not None


def test_base_model():
    db.connect(DB_URI)

    class User(db.BaseModel):
        __tablename__ = "users"
        id = Column(Integer, primary_key=True)

    db.create_all()

    # using the exists columns
    users = db.query(User.created_time, User.updated_time, User.is_active).all()
    assert users == []


def test_base_model_create_entity():
    db.connect(DB_URI)

    class User(db.BaseModel):
        __tablename__ = "users"
        id = Column(Integer, primary_key=True)

    db.create_all()
    user = User.create()
    assert user.id > 0  # create successfully
    assert user.created_time <= user.updated_time
    assert user.is_active
